import cf_units
import numpy as np

from pyaerocom import const
from pyaerocom.variable_helpers import get_variable


def calc_ang4487aer(data):
    """Compute Angstrom coefficient (440-870nm) from 440 and 870 nm AODs

    Parameters
    ----------
    data : dict-like
        data object containing imported results

    Note
    ----
    Requires the following two variables to be available in provided data
    object:

        1. od440aer
        2. od870aer

    Raises
    ------
    AttributError
        if either 'od440aer' or 'od870aer' are not available in data object

    Returns
    -------
    ndarray
        array containing computed angstrom coefficients


    """
    if not all([x in data for x in ["od440aer", "od870aer"]]):
        raise AttributeError(
            "Either of the two (or both) required variables "
            "(od440aer, od870aer) are not available in data"
        )
    od440aer, od870aer = data["od440aer"], data["od870aer"]
    return compute_angstrom_coeff(od440aer, od870aer, 0.44, 0.87)


def calc_od550aer(data):
    """Compute AOD at 550 nm using Angstrom coefficient and 500 nm AOD

    Parameters
    ----------
    data : dict-like
        data object containing imported results

    Returns
    -------
    :obj:`float` or :obj:`ndarray`
        AOD(s) at shifted wavelength
    """
    return _calc_od_helper(
        data=data,
        var_name="od550aer",
        to_lambda=0.55,
        od_ref="od500aer",
        lambda_ref=0.50,
        od_ref_alt="od440aer",
        lambda_ref_alt=0.44,
        use_angstrom_coeff="ang4487aer",
    )


def calc_abs550aer(data):
    """Compute AOD at 550 nm using Angstrom coefficient and 500 nm AOD

    Parameters
    ----------
    data : dict-like
        data object containing imported results

    Returns
    -------
    :obj:`float` or :obj:`ndarray`
        AOD(s) at shifted wavelength
    """
    return _calc_od_helper(
        data=data,
        var_name="abs550aer",
        to_lambda=0.55,
        od_ref="abs500aer",
        lambda_ref=0.50,
        od_ref_alt="abs440aer",
        lambda_ref_alt=0.44,
        use_angstrom_coeff="angabs4487aer",
    )


def calc_od550gt1aer(data):
    """Compute coarse mode AOD at 550 nm using Angstrom coeff. and 500 nm AOD

    Parameters
    ----------
    data : dict-like
        data object containing imported results

    Returns
    -------
    float or ndarray
        AOD(s) at shifted wavelength
    """
    return _calc_od_helper(
        data=data,
        var_name="od550gt1aer",
        to_lambda=0.55,
        od_ref="od500gt1aer",
        lambda_ref=0.50,
        use_angstrom_coeff="ang4487aer",
    )


def calc_od550lt1aer(data):
    """Compute fine mode AOD at 550 nm using Angstrom coeff. and 500 nm AOD

    Parameters
    ----------
    data : dict-like
        data object containing imported results

    Returns
    -------
    :obj:`float` or :obj:`ndarray`
        AOD(s) at shifted wavelength
    """
    return _calc_od_helper(
        data=data,
        var_name="od550lt1aer",
        to_lambda=0.55,
        od_ref="od500lt1aer",
        lambda_ref=0.50,
        use_angstrom_coeff="ang4487aer",
    )


def calc_od550lt1ang(data):
    """Compute AOD at 550 nm using Angstrom coeff. and 500 nm AOD,
        that is filtered for angstrom coeff < 1 to get AOD representative
        of coarse particles.

    Parameters
    ----------
    data : dict-like
        data object containing imported results

    Returns
    -------
    :obj:`float` or :obj:`ndarray`
        AOD(s) at shifted wavelength
    """

    return _calc_od_helper(
        data=data,
        var_name="od550lt1ang",
        to_lambda=0.55,
        od_ref="od500aer",
        lambda_ref=0.50,
        od_ref_alt="od440aer",
        lambda_ref_alt=0.44,
        use_angstrom_coeff="ang4487aer",
        treshold_angstrom=1.0,
    )


def compute_angstrom_coeff(od1, od2, lambda1, lambda2):
    """Compute Angstrom coefficient based on 2 optical densities

    Parameters
    ----------
    od1 : :obj:`float` or :obj:`ndarray`
        AOD at wavelength 1
    od2 : :obj:`float` or :obj:`ndarray`
        AOD at wavelength 2
    lambda1 : :obj:`float` or :obj:`ndarray`
        wavelength 1
    lambda 2 : :obj:`float` or :obj:`ndarray`
        wavelength 2

    Returns
    -------
    :obj:`float` or :obj:`ndarray`
        Angstrom exponent(s)
    """
    return -np.log(od1 / od2) / np.log(lambda1 / lambda2)


def compute_od_from_angstromexp(to_lambda, od_ref, lambda_ref, angstrom_coeff):
    """Compute AOD at specified wavelength

    Uses Angstrom coefficient and reference AOD to compute the
    corresponding wavelength shifted AOD

    Parameters
    ----------
    to_lambda : :obj:`float` or :obj:`ndarray`
        wavelength for which AOD is calculated
    od_ref : :obj:`float` or :obj:`ndarray`
        reference AOD
    lambda_ref : :obj:`float` or :obj:`ndarray`
        wavelength corresponding to reference AOD
    angstrom_coeff : :obj:`float` or :obj:`ndarray`
        Angstrom coefficient

    Returns
    -------
    :obj:`float` or :obj:`ndarray`
        AOD(s) at shifted wavelength

    """
    return od_ref * (lambda_ref / to_lambda) ** angstrom_coeff


def _calc_od_helper(
    data,
    var_name,
    to_lambda,
    od_ref,
    lambda_ref,
    od_ref_alt=None,
    lambda_ref_alt=None,
    use_angstrom_coeff="ang4487aer",
    treshold_angstrom=None,
):
    """Helper method for computing ODs

    Parameters
    ----------
    data : dict-like
        data object containing loaded results used to compute the ODs at a new
        wavelength
    var_name : str
        name of variable that is supposed to be computed (is used in order to
        see whether a global lower threshold is defined for this variable and
        if this is the case, all computed values that are below this threshold
        are replaced with NaNs)
    to_lambda : float
        wavelength of computed AOD
    od_ref : str
        name of reference AOD in data
    lambda_ref : :obj:`float` or :obj:`ndarray`
        wavelength corresponding to reference AOD
    od_ref_alt : str
        alternative reference AOD (is used for datapoints where former is
        invalid)
    lambda_ref_alt : :obj:`float` or :obj:`ndarray`, optional
        wavelength corresponding to alternative reference AOD
    use_angstrom_coeff : str
        name of Angstrom coefficient in data, that is used for computation
    threshold_angstrom : float
        filter out observations that have angstrom exponent larger than a set
        threshold.

    Returns
    -------
    :obj:`float` or :obj:`ndarray`
        AOD(s) at shifted wavelength

    Raises
    ------
    AttributeError
        if neither ``od_ref`` nor ``od_ref_alt`` are available in data, or if
        ``use_angstrom_coeff`` is missing
    """
    if not od_ref in data:
        if od_ref_alt is None or not od_ref_alt in data:
            raise AttributeError(f"No alternative OD found for computation of {var_name}")
        return compute_od_from_angstromexp(
            to_lambda=to_lambda,
            od_ref=data[od_ref_alt],
            lambda_ref=lambda_ref_alt,
            angstrom_coeff=data[use_angstrom_coeff],
        )
    elif not use_angstrom_coeff in data:
        raise AttributeError("Angstrom coefficient (440-870 nm) is not available in provided data")
    result = compute_od_from_angstromexp(
        to_lambda=to_lambda,
        od_ref=data[od_ref],
        lambda_ref=lambda_ref,
        angstrom_coeff=data[use_angstrom_coeff],
    )

    # optional if available
    if od_ref_alt in data:
        # fill up time steps that are nans with values calculated from the
        # alternative wavelength to minimise gaps in the time series
        mask = np.argwhere(np.isnan(result))

        if len(mask) > 0:  # there are nans
            ods_alt = data[od_ref_alt][mask]
            ang = data[use_angstrom_coeff][mask]
            replace = compute_od_from_angstromexp(
                to_lambda=to_lambda, od_ref=ods_alt, lambda_ref=lambda_ref_alt, angstrom_coeff=ang
            )
            result[mask] = replace
    if treshold_angstrom:
        result = np.where(data[use_angstrom_coeff] < treshold_angstrom, result, np.nan)

    return result


def compute_ang4470dryaer_from_dry_scat(data):
    """Compute angstrom exponent between 440 and 700 nm

    Parameters
    ----------
    StationData or dict
        data containing dry scattering coefficients at 440 and 700 nm
        (i.e. keys sc440dryaer and sc700dryaer)

    Returns
    -------
    StationData or dict
        extended data object containing angstrom exponent
    """
    return compute_angstrom_coeff(data["sc440dryaer"], data["sc700dryaer"], 440, 700)


def compute_sc550dryaer(data):
    """Compute dry scattering coefficent applying RH threshold

    Cf. :func:`_compute_dry_helper`

    Parameters
    ----------
    dict
        data object containing scattering and RH data

    Returns
    -------
    dict
        modified data object containing new column sc550dryaer

    """
    rh_max = get_variable("sc550dryaer").dry_rh_max
    vals, rh_mean = _compute_dry_helper(
        data, data_colname="sc550aer", rh_colname="scrh", rh_max_percent=rh_max
    )
    if not "sc550dryaer" in data.var_info:
        data.var_info["sc550dryaer"] = {}
    data.var_info["sc550dryaer"]["rh_mean"] = rh_mean

    return vals


def compute_sc440dryaer(data):
    """Compute dry scattering coefficent applying RH threshold

    Cf. :func:`_compute_dry_helper`

    Parameters
    ----------
    dict
        data object containing scattering and RH data

    Returns
    -------
    dict
        modified data object containing new column sc550dryaer

    """
    rh_max = get_variable("sc440dryaer").dry_rh_max
    return _compute_dry_helper(
        data, data_colname="sc440aer", rh_colname="scrh", rh_max_percent=rh_max
    )[0]


def compute_sc700dryaer(data):
    """Compute dry scattering coefficent applying RH threshold

    Cf. :func:`_compute_dry_helper`

    Parameters
    ----------
    dict
        data object containing scattering and RH data

    Returns
    -------
    dict
        modified data object containing new column sc550dryaer

    """
    rh_max = get_variable("sc700dryaer").dry_rh_max
    return _compute_dry_helper(
        data, data_colname="sc700aer", rh_colname="scrh", rh_max_percent=rh_max
    )[0]


def compute_ac550dryaer(data):
    """Compute aerosol dry absorption coefficent applying RH threshold

    Cf. :func:`_compute_dry_helper`

    Parameters
    ----------
    dict
        data object containing scattering and RH data

    Returns
    -------
    dict
        modified data object containing new column sc550dryaer

    """
    rh_max = get_variable("ac550dryaer").dry_rh_max
    return _compute_dry_helper(
        data, data_colname="ac550aer", rh_colname="acrh", rh_max_percent=rh_max
    )[0]


def _compute_dry_helper(data, data_colname, rh_colname, rh_max_percent=None):
    """Compute new column that contains data where RH is smaller than ...

    All values in original data columns are set to NaN, where RH exceeds a
    certain threshold or where RH is NaN.

    Parameters
    ----------
    data : dict-like
        dictionary-like object that contains data
    data_colname : str
        column name of variable data that is supposed to be filtered
    rh_colname : str
        column name of RH data
    rh_max_percent : int
        maximum relative humidity

    Returns
    -------
    dict
        modified data dictionary with new dry data column
    """
    if rh_max_percent is None:
        rh_max_percent = const.RH_MAX_PERCENT_DRY

    vals = np.array(data[data_colname], copy=True)

    rh = data[rh_colname]

    high_rh = rh > rh_max_percent
    rhnan = np.isnan(rh)
    vals[high_rh] = np.nan
    vals[rhnan] = np.nan

    rh = rh[~rhnan]
    lowrh = rh[~high_rh[~rhnan]]
    if len(lowrh) > 0:
        rh_mean = np.nanmean(lowrh)
    else:
        rh_mean = np.nan

    return vals, rh_mean


def _compute_wdep_from_concprcp_helper(data, wdep_var, concprcp_var, pr_var):
    """
    Helper to compute wed deposition from concentration in precipitation

    Note
    ----
    In addition to the returned numpy array, the input instance of
    :class:`StationData` is modified by additional metadata and flags for
    the new variable.

    Parameters
    ----------
    data : StationData
        input data containing concentration in precipitation variable and
        precipitation variable. Both are needed to be sampled at the same
        time and both arrays must have the same lengths.
    wdep_var : str
        name of output wet deposition variable
    concprcp_var : str
        name of concenration in precipitation variable
    pr_var : str
        precipitation variable (needs to be implicit, i.e. in units of mm or
        similar and not in units of mm d-1.

    Returns
    -------
    numpy.ndarray
        array with wet deposition values
    """

    vars_needed = (concprcp_var, pr_var)

    if not all(x in data.data_flagged for x in vars_needed):
        raise ValueError(f"Need flags for {vars_needed} to compute wet deposition")
    from pyaerocom import TsType
    from pyaerocom.units_helpers import RATES_FREQ_DEFAULT, get_unit_conversion_fac

    tst = TsType(data.get_var_ts_type(concprcp_var))

    ival = tst.to_si()

    conc_unit = data.get_unit(concprcp_var)
    conc_data = data[concprcp_var]
    if not conc_unit.endswith("m-3"):
        raise NotImplementedError("Can only handle concprcp unit ending with m-3")
    concprcp_flags = data.data_flagged[concprcp_var]

    pr_unit = data.get_unit(pr_var)
    if not pr_unit == "m":
        data.convert_unit(pr_var, "m")
    pr_data = data[pr_var]
    pr_flags = data.data_flagged[pr_var]
    # check where precip data is zero (it did not rain!) and for each of
    # these timestamps, set concprcp to 0 (it should be 0 if there is no
    # rain...) and set flags in concprcp to False (these data are to be used
    # later)
    pr_zero = pr_data == 0
    if pr_zero.sum() > 0:
        conc_data[pr_zero] = 0
        concprcp_flags[pr_zero] = False
        pr_flags[pr_zero] = False
    wdep = conc_data * pr_data
    wdep_units = conc_unit.replace("m-3", "m-2")

    if not ival == RATES_FREQ_DEFAULT:
        fac = get_unit_conversion_fac(ival, RATES_FREQ_DEFAULT)
        wdep /= fac
    # in units of ts_type, that is, e.g. kg m-2 d
    freq_str = f" {RATES_FREQ_DEFAULT}-1"
    wdep_units += freq_str
    if not wdep_var in data.var_info:
        data.var_info[wdep_var] = {}
    data.var_info[wdep_var]["units"] = wdep_units

    # set flags for wetso4
    wdep_flags = np.zeros(len(wdep)).astype(bool)
    wdep_flags[concprcp_flags] = True
    wdep_flags[pr_flags] = True
    data.data_flagged[wdep_var] = wdep_flags

    return wdep


def compute_wetoxs_from_concprcpoxs(data):
    """Compute wdep from conc in precip and precip data

    Note
    ----
    In addition to the returned numpy array, the input instance of
    :class:`StationData` is modified by additional metadata and flags for
    the new variable. See also :func:`_compute_wdep_from_concprcp_helper`.

    Parameters
    ----------
    StationData
        data object containing concprcp and precip data

    Returns
    -------
    numpy.ndarray
        array with wet deposition values

    """
    return _compute_wdep_from_concprcp_helper(data, "wetoxs", "concprcpoxs", "pr")


def compute_wetoxn_from_concprcpoxn(data):
    """Compute wdep from conc in precip and precip data

    Note
    ----
    In addition to the returned numpy array, the input instance of
    :class:`StationData` is modified by additional metadata and flags for
    the new variable. See also :func:`_compute_wdep_from_concprcp_helper`.

    Parameters
    ----------
    StationData
        data object containing concprcp and precip data

    Returns
    -------
    numpy.ndarray
        array with wet deposition values

    """
    return _compute_wdep_from_concprcp_helper(data, "wetoxn", "concprcpoxn", "pr")


def compute_wetrdn_from_concprcprdn(data):
    """Compute wdep from conc in precip and precip data

    Note
    ----
    In addition to the returned numpy array, the input instance of
    :class:`StationData` is modified by additional metadata and flags for
    the new variable. See also :func:`_compute_wdep_from_concprcp_helper`.

    Parameters
    ----------
    StationData
        data object containing concprcp and precip data

    Returns
    -------
    numpy.ndarray
        array with wet deposition values

    """
    return _compute_wdep_from_concprcp_helper(data, "wetrdn", "concprcprdn", "pr")


def vmrx_to_concx(data, p_pascal, T_kelvin, vmr_unit, mmol_var, mmol_air=None, to_unit=None):
    """
    Convert volume mixing ratio (vmr) to mass concentration

    Parameters
    ----------
    data : float or ndarray
        array containing vmr values
    p_pascal : float
        pressure in Pa of input data
    T_kelvin : float
        temperature in K of input data
    vmr_unit : str
        unit of input data
    mmol_var : float
        molar mass of variable represented by input data
    mmol_air : float, optional
        Molar mass of air. Uses average density of dry air if None.
        The default is None.
    to_unit : str, optional
        Unit to which output data is converted. If None, output unit is
        kg m-3. The default is None.

    Returns
    -------
    float or ndarray
        input data converted to mass concentration

    """
    if mmol_air is None:
        from pyaerocom.molmasses import get_molmass

        mmol_air = get_molmass("air_dry")

    Rspecific = 287.058  # J kg-1 K-1

    conversion_fac = 1 / cf_units.Unit("mol mol-1").convert(1, vmr_unit)

    airdensity = p_pascal / (Rspecific * T_kelvin)  # kg m-3
    mulfac = mmol_var / mmol_air * airdensity  # kg m-3
    conc = data * mulfac  # kg m-3
    if to_unit is not None:
        conversion_fac *= cf_units.Unit("kg m-3").convert(1, to_unit)
    if not np.isclose(conversion_fac, 1, rtol=1e-7):
        conc *= conversion_fac
    return conc


def concx_to_vmrx(data, p_pascal, T_kelvin, conc_unit, mmol_var, mmol_air=None, to_unit=None):
    """
    Convert mass concentration to volume mixing ratio (vmr)

    Parameters
    ----------
    data : float or ndarray
        array containing vmr values
    p_pascal : float
        pressure in Pa of input data
    T_kelvin : float
        temperature in K of input data
    vmr_unit : str
        unit of input data
    mmol_var : float
        molar mass of variable represented by input data
    mmol_air : float, optional
        Molar mass of air. Uses average density of dry air if None.
        The default is None.
    to_unit : str, optional
        Unit to which output data is converted. If None, output unit is
        kg m-3. The default is None.

    Returns
    -------
    float or ndarray
        input data converted to volume mixing ratio

    """
    if mmol_air is None:
        from pyaerocom.molmasses import get_molmass

        mmol_air = get_molmass("air_dry")

    Rspecific = 287.058  # J kg-1 K-1

    conversion_fac = 1 / cf_units.Unit("kg m-3").convert(1, conc_unit)

    airdensity = p_pascal / (Rspecific * T_kelvin)  # kg m-3
    mulfac = mmol_var / mmol_air * airdensity  # kg m-3
    vmr = data / mulfac  # unitless
    if to_unit is not None:
        conversion_fac *= cf_units.Unit("mole mole-1").convert(1, to_unit)
    if not np.isclose(conversion_fac, 1, rtol=1e-7):
        vmr *= conversion_fac
    return vmr


def calc_vmro3max(data):

    var_name = "vmro3"
    new_var_name = "vmro3max"

    flags = data.data_flagged[var_name]

    o3max = data[var_name]

    units = data.var_info[var_name]["units"]
    # data.var_info[new_var_name]["units"] = units

    if not new_var_name in data.var_info:
        data.var_info[new_var_name] = {}

    data.var_info[new_var_name] = data.var_info[var_name]

    data.data_flagged[new_var_name] = flags
    # print(data.var_info)
    # exit()
    return o3max



import os

import numpy as np
import pandas as pd
from geonum import atmosphere as atm
from tqdm import tqdm

from pyaerocom import const
from pyaerocom.aux_var_helpers import concx_to_vmrx
from pyaerocom.exceptions import DataRetrievalError
from pyaerocom.helpers import varlist_aerocom
from pyaerocom.io import ReadUngriddedBase
from pyaerocom.molmasses import get_molmass
from pyaerocom.ungriddeddata import UngriddedData

P_STD = atm.p0  # standard atmosphere pressure
T_STD = atm.T0_STD  # standard atmosphere temperature


def _conc_to_vmr_marcopolo_stats(
    data, to_var, from_var, p_pascal=None, T_kelvin=None, mmol_air=None
):
    if p_pascal is None:
        p_pascal = P_STD
    if T_kelvin is None:
        T_kelvin = T_STD
    if mmol_air is None:
        mmol_air = get_molmass("air_dry")

    for stat in data:
        if not from_var in stat:
            continue

        concdata = stat[from_var]

        mmol_var = get_molmass(to_var)
        from_unit = stat["var_info"][from_var]["units"]
        to_unit = const.VARS[to_var].units
        vmrvals = concx_to_vmrx(
            concdata,
            p_pascal=p_pascal,
            T_kelvin=T_kelvin,
            mmol_var=mmol_var,
            mmol_air=mmol_air,
            conc_unit=from_unit,
            to_unit=to_unit,
        )
        stat[to_var] = vmrvals
        vi = {}
        vi.update(stat["var_info"][from_var])
        vi["computed"] = True
        vi["units"] = to_unit
        vi["units_info"] = (
            f"The original data is provided as mass conc. in units of ug m-3 and "
            f"was converted to ppb assuming a standard atmosphere "
            f"(p={p_pascal/100}hPa, T={T_kelvin}K) assuming dry molar mass of air "
            f"M_Air_dry={mmol_air}g/mol and total molecular mass of "
            f"{from_var}={mmol_var}g/mol"
        )
        stat["var_info"][to_var] = vi

    return data


class ReadMarcoPolo(ReadUngriddedBase):
    """
    Reading routine for Chinese MarcoPolo observations
    """

    #: Data type of files
    _FILETYPE = ".csv"

    #: Filemask for glob data file search
    _FILEMASK = f"*{_FILETYPE}"

    #: Column delimiter
    FILE_COL_DELIM = ","

    #: Version log of this class (for caching)
    __version__ = "0.02"

    #: Default data ID
    DATA_ID = "MarcoPolo"

    #: List of all datasets supported by this interface
    SUPPORTED_DATASETS = [DATA_ID]

    #: Indices of columns in data files
    FILE_COLS = {"station_id": 0, "datetime": 1, "var_name": 2, "value": 3}
    #: Mapping of columns in station metadata file to pyaerocom standard
    STATION_META_MAP = {
        "stationId": "station_id",
        "stationName": "station_name",
        "cityId": "city_id",
        "latitude": "latitude",
        "longitude": "longitude",
        "cityName": "city",
    }

    #: conversion functions for metadata dtypes
    STATION_META_DTYPES = {
        "station_id": str,
        "station_name": str,
        "city_id": str,
        "latitude": float,
        "longitude": float,
        "city": str,
    }

    #: Variable names in data files
    VAR_MAP = {
        "concco": "co",
        "concso2": "so2",
        "concno2": "no2",
        "conco3": "o3",
        "concpm10": "pm10",
        "concpm25": "pm2_5",
    }

    #: Variable units (not provided in data files)
    VAR_UNITS = {
        "concco": "unknown",
        "concso2": "unknown",
        "concno2": "ug m-3",  # mail from Henk Eskes, 1 Feb. 2021
        "conco3": "ug m-3",  # mail from Henk Eskes, 1 Feb. 2021
        "concpm10": "ug m-3",  # mail from Henk Eskes, 1 Feb. 2021
        "concpm25": "ug m-3",  # mail from Henk Eskes, 1 Feb. 2021
    }

    #: Variables that are computed (cannot be read directly)
    AUX_REQUIRES = {"vmro3": ["conco3"], "vmrno2": ["concno2"]}

    #: functions used to convert variables that are computed
    AUX_FUNS = {"vmro3": _conc_to_vmr_marcopolo_stats, "vmrno2": _conc_to_vmr_marcopolo_stats}

    PROVIDES_VARIABLES = list(VAR_MAP)
    DEFAULT_VARS = ["concpm10", "concpm25", "concno2", "conco3"]

    TS_TYPE = "hourly"

    STAT_METADATA_FILENAME = "station_info.xlsx"

    def __init__(self, data_id=None, data_dir=None):
        super().__init__(data_id=data_id, data_dir=data_dir)

        try:
            import openpyxl
        except ImportError:
            raise ImportError(
                "ReadMarcoPolo class needs package openpyxl "
                "which is not part of the pyaerocom standard installation. "
                "Please install openpyxl via conda or pip."
            )

    def _read_metadata_file(self):
        fn = os.path.join(self.data_dir, self.STAT_METADATA_FILENAME)
        cfg = pd.read_excel(fn, engine="openpyxl")
        return cfg

    def _init_station_metadata(self):
        """
        Initiate metadata for all stations

        Returns
        -------
        dict
            dictionary with metadata dictionaries for all stations

        """

        cfg = self._read_metadata_file()

        meta_map = self.STATION_META_MAP
        dtypes = self.STATION_META_DTYPES

        cols = list(cfg.columns.values)
        col_idx = {}
        for from_meta, to_meta in meta_map.items():
            col_idx[to_meta] = cols.index(from_meta)

        arr = cfg.values

        stats = {}
        for row in arr:
            stat = {}
            for meta_key, col_num in col_idx.items():
                stat[meta_key] = dtypes[meta_key](row[col_num])
            stat["data_id"] = self.data_id
            stat["ts_type"] = self.TS_TYPE
            stats[stat["station_id"]] = stat

        return stats

    def _read_file(self, file):
        df = pd.read_csv(file, sep=self.FILE_COL_DELIM)
        return df

    def _make_station_data(
        self, var, var_stats_unique, stat_ids, var_stats, var_dtime, var_values, stat_meta
    ):

        units = self.VAR_UNITS
        stats = []
        for stat_id in var_stats_unique:
            if not stat_id in stat_ids:
                continue

            statmask = var_stats == stat_id

            if statmask.sum() == 0:
                continue

            timestamps = var_dtime[statmask]
            vals = var_values[statmask]

            stat = {}
            stat.update(stat_meta[stat_id])

            stat["dtime"] = timestamps
            stat["timezone"] = "unknown"
            stat[var] = vals
            unit = units[var]
            stat["var_info"] = {}
            stat["var_info"][var] = dict(units=unit)
            stats.append(stat)
        return stats

    def _read_files(self, files, vars_to_retrieve):

        filecols = self.FILE_COLS

        stat_meta = self._init_station_metadata()

        # get all station IDs found in the metadata file
        stat_ids = list(stat_meta)

        arrs = []

        varcol = filecols["var_name"]
        for i in tqdm(range(len(files))):
            filedata = self._read_file(files[i])
            arr = filedata.values

            for i, var in enumerate(vars_to_retrieve):
                filevar = self.VAR_MAP[var]
                if i == 0:
                    mask = arr[:, varcol] == filevar
                else:
                    mask = np.logical_or(mask, arr[:, varcol] == filevar)
            matches = mask.sum()
            if matches:
                vardata = arr[mask]
                arrs.append(vardata)
        if len(arrs) == 0:
            raise DataRetrievalError("None of the input variables could be found in input list")
        data = np.concatenate(arrs)

        dtime = data[:, filecols["datetime"]].astype("datetime64[s]")

        varis = data[:, varcol].astype(str)
        values = data[:, filecols["value"]].astype(np.float64)
        statids = data[:, filecols["station_id"]].astype(str)

        stats = []
        for var in vars_to_retrieve:
            var_in_file = self.VAR_MAP[var]

            varmask = varis == var_in_file

            var_values = values[varmask]
            var_stats = statids[varmask]
            var_dtime = dtime[varmask]

            var_stats_unique = np.unique(var_stats)

            var_stats = self._make_station_data(
                var, var_stats_unique, stat_ids, var_stats, var_dtime, var_values, stat_meta
            )
            stats.extend(var_stats)

        return stats

    def read_file(self):
        """
        This method is not implemented (but needs to be declared for template)

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError("Not needed for these data since the format is unsuitable...")

    def compute_additional_vars(self, statlist_from_file, vars_to_compute):

        for var in vars_to_compute:
            fun = self.AUX_FUNS[var]
            requires = self.AUX_REQUIRES[var]
            # this will add the variable data to each station data in
            # statlist_from_file
            statlist_from_file = fun(statlist_from_file, var, *requires)

        return statlist_from_file

    def read(self, vars_to_retrieve=None, first_file=None, last_file=None):
        """
        Read variable data

        Parameters
        ----------
        vars_to_retrieve : str or list, optional
            List of variables to be retrieved. The default is None.
        first_file : int, optional
            Index of first file to be read. The default is None, in which case
            index 0 in file list is used.
        last_file : int, optional
            Index of last file to be read. The default is None, in which case
            last index in file list is used.

        Returns
        -------
        data : UngriddedData
            loaded data object.

        """
        if vars_to_retrieve is None:
            vars_to_retrieve = self.DEFAULT_VARS
        elif isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]

        # make sure to use AeroCom variable names in output data
        vars_to_retrieve = varlist_aerocom(vars_to_retrieve)

        vars_to_read, vars_to_compute = self.check_vars_to_retrieve(vars_to_retrieve)

        files = self.get_file_list()
        if first_file is None:
            first_file = 0
        if last_file is None:
            last_file = len(files)
        files = files[first_file:last_file]

        stats = self._read_files(files, vars_to_read)

        stats = self.compute_additional_vars(stats, vars_to_compute)

        data = UngriddedData.from_station_data(stats)

        return data



"""
Logging configuration and package metadata helpers

NOTE: 
All pyaerocom child modules share the logging configuration
- all logging messages are time stamped and writen out to file
- some messages are also printed to the console
- log files are kept up to 14 days
- logging configuration is read from pyaerocom/data/logging.ini
  with default values from LOGGING_CONFIG
"""
from __future__ import annotations

import logging
from importlib import resources
from logging.config import fileConfig

LOGGING_CONFIG = dict(
    # root logger
    file_name="pyaerocom.log",
    file_days=14,
    file_level="DEBUG",
    # pyaerocom logger
    console_level="INFO",
)

with resources.path("pyaerocom.data", "logging.ini") as path:
    fileConfig(path, defaults=LOGGING_CONFIG, disable_existing_loggers=False)


def change_verbosity(level: str | int) -> None:
    """
    Change logging verbosity (to console)

    Parameters
    ----------
    level: str or int
        new `logging level<https://docs.python.org/3/library/logging.html#logging-levels>`_

    Returns
    -------
    None

    """
    if isinstance(level, str):
        level = level.upper()

    if isinstance(level, int) and not (logging.DEBUG <= level <= logging.CRITICAL):
        raise ValueError(
            f"invalid logging level {level}, choose a value between {logging.DEBUG} and {logging.CRITICAL}"
        )

    logger = logging.getLogger(__package__)
    assert logger.handlers, f"{logger.name} logger has not been configured correctly"
    for handler in logger.handlers:
        if type(handler) == logging.StreamHandler:
            handler.setLevel(level)



from typing import Optional

import typer

from pyaerocom import __package__, __version__, const, tools

main = typer.Typer()


def version_callback(value: bool):
    if not value:
        return

    typer.echo(f"{__package__} {__version__}")
    raise typer.Exit()


@main.callback()
def callback(
    version: Optional[bool] = typer.Option(None, "--version", "-V", callback=version_callback)
):
    """Pyaerocom Command Line Interface"""


@main.command()
def browse(database: str = typer.Argument(..., help="Provide database name.")):
    """Browse database e.g., browse <DATABASE>"""
    print(f"Searching database for matches of {database}")
    print(tools.browse_database(database))


@main.command()
def clearcache():
    """Delete cached data objects"""

    delete = typer.confirm("Are you sure you want to delete all cached data objects?")
    if delete:
        print("OK then.... here we go!")
        tools.clear_cache()
    else:
        print("Wise decision, pyaerocom will handle it for you automatically anyways ;P")


@main.command()
def ppiaccess():
    """Check if MetNO PPI can be accessed"""
    print("True") if const.has_access_lustre else print("False")


if __name__ == "__main__":
    main()



import logging
import os
import sqlite3

from pyaerocom._lowlevel_helpers import BrowseDict

logger = logging.getLogger(__name__)


class EbasSQLRequest(BrowseDict):
    """Low level dictionary like object for EBAS sqlite queries

    Attributes
    ----------
    variables : :obj:`tuple`, optional
        tuple containing variable names to be extracted (e.g.
        ``('aerosol_light_scattering_coefficient', 'aerosol_optical_depth')``).
        If None, all available is used
    start_date : :obj:`str`, optional
        start date of data request (format YYYY-MM-DD). If None, all available
        is used
    stop_date : :obj:`str`, optional
        stop date of data request (format YYYY-MM-DD). If None, all available
        is used
    station_names : :obj:`tuple`, optional
        tuple containing station_names of request (e.g.
        ``('Birkenes II', 'Asa')``).If None, all available is used
    matrices : :obj:`tuple`, optional
        tuple containing station_names of request (e.g.
        ``('pm1', 'pm10', 'pm25', 'aerosol')``)
        If None, all available is used
    altitude_range : :obj:`tuple`, optional
        tuple specifying altitude range of station in m (e.g.
        ``(0.0, 500.0)``). If None, all available is used
    lon_range : :obj:`tuple`, optional
        tuple specifying longitude range of station in degrees (e.g.
        ``(-20, 20)``). If None, all available is used
    lat_range : :obj:`tuple`, optional
        tuple specifying latitude range of station in degrees (e.g.
        ``(50, 80)``). If None, all available is used
    instrument_type : :obj:`str`, optional
        string specifying instrument types (e.g.
        ``("nephelometer")``)
    statistics : :obj:`tuple`, optional
        string specifying statistics code (e.g.
        ``("arithmetic mean")``)

    Parameters
    ----------
    see Attributes
    """

    def __init__(
        self,
        variables=None,
        start_date=None,
        stop_date=None,
        station_names=None,
        matrices=None,
        altitude_range=None,
        lon_range=None,
        lat_range=None,
        instrument_types=None,
        statistics=None,
        datalevel=None,
    ):
        self.variables = variables
        self.start_date = start_date
        self.stop_date = stop_date
        self.station_names = station_names
        self.matrices = matrices
        self.altitude_range = altitude_range
        self.lon_range = lon_range
        self.lat_range = lat_range
        self.instrument_types = instrument_types
        self.statistics = statistics
        self.datalevel = datalevel

    def update(self, **kwargs):
        for k, v in kwargs.items():
            if k in self:
                self[k] = v
            else:
                logger.warning(f"Unknown EBAS SQL request key {k} (value {v})")

    @staticmethod
    def _var2sql(var):
        if isinstance(var, list):
            if len(var) > 1:
                var = tuple(var)
            else:
                var = var[0]

        if isinstance(var, tuple):
            return f"{var}"
        elif isinstance(var, str):
            return f"('{var}')"
        raise ValueError("Invalid value...")

    def make_file_query_str(self, distinct=True, **kwargs):
        """Wrapper for base method :func:`make_query_str`

        Parameters
        ----------
        distinct : bool
            return unique files
        **kwargs
            update request attributes (e.g. ``lon_range=(30, 60)``)

        Returns
        -------
        str
            SQL file request command for current specs
        """
        query = self.make_query_str(distinct=distinct, **kwargs)
        # add an extsion to get only files that have no fraction variables in them
        query = query.replace(
            ";",
            " and not exists (select * from characteristic where var_id=variable.var_id and ct_type='Fraction');",
        )
        return query

    def make_query_str(self, what=None, distinct=True, **kwargs):
        """Translate current class state into SQL query command string

        Parameters
        ----------
        what : str or tuple, optional
            what columns to retrieve (e.g. comp_name for all variables) from
            table specified. Defaults to None, in which case "filename" is used
        distinct : bool
            return unique files
        **kwargs
            update request attributes (e.g. ``lon_range=(30, 60)``)

        Returns
        -------
        str
            SQL file request command for current specs
        """
        self.update(**kwargs)
        if what is None:
            what = "filename"
        elif not isinstance(what, str):  # tuple or list of parameters to be retrieved
            what = ",".join(what)

        if distinct:
            req = f"select distinct {what} from variable"
        else:
            req = f"select {what} from variable"
        req += " join station on station.station_code=variable.station_code"
        add_cond = 0
        # add constraints from station table
        conv = self._var2sql
        if self.station_names is not None:
            req += f" where station_name in {conv(self.station_names)}"
            add_cond += 1
        if self.altitude_range is not None:
            low, high = self.altitude_range
            req += " and " if add_cond else " where "
            req += f"station_altitude>{low} and station_altitude<{high}"
            add_cond += 1
        if self.lon_range is not None:
            l, r = self.lon_range
            req += " and " if add_cond else " where "
            req += f"station_longitude>{l} and station_longitude<{r}"
            add_cond += 1
        if self.lat_range is not None:
            s, n = self.lat_range
            req += " and " if add_cond else " where "
            req += f"station_latitude>{s} and station_latitude<{n}"
            add_cond += 1
        if self.instrument_types is not None:
            req += " and " if add_cond else " where "
            req += f"instr_type in {conv(self.instrument_types)}"
            add_cond += 1
        # add constraints from variable table
        if self.variables is not None:
            req += " and " if add_cond else " where "
            req += f"comp_name in {conv(self.variables)}"
            add_cond += 1
        if self.stop_date is not None:
            req += " and " if add_cond else " where "
            req += f"first_end < '{self.stop_date}'"
            add_cond += 1
        if self.start_date is not None:
            req += " and " if add_cond else " where "
            req += f"last_start > '{self.start_date}'"
            add_cond += 1
        if self.matrices is not None:
            req += " and " if add_cond else " where "
            req += f"matrix in {conv(self.matrices)}"
            add_cond += 1
        if self.statistics is not None:
            req += " and " if add_cond else " where "
            req += f"statistics in {conv(self.statistics)}"
            add_cond += 1
        if self.datalevel is not None:
            req += " and " if add_cond else " where "
            req += f"datalevel={self.datalevel}"
            add_cond += 1
        return req + ";"

    def __str__(self):
        head = f"Pyaerocom {type(self).__name__}"
        s = f"\n{head}\n{len(head) * '-'}"
        for k, v in self.items():
            s += f"\n{k}: {v}"
        s += f"\nFilename request string:\n{self.make_file_query_str()}"
        return s


class EbasFileIndex:
    """EBAS SQLite I/O interface

    Takes care of connection to database and execution of requests
    """

    def __init__(self, database=None):
        self._database = database

    @property
    def database(self):
        """Path to ebas_file_index.sqlite3 file"""
        db = self._database
        if db is None or not os.path.exists(db):
            raise AttributeError(
                "EBAS SQLite database file could not be "
                "located but is needed in EbasFileIndex class"
            )
        return db

    @property
    def ALL_STATION_NAMES(self):
        """List of all available station names in database"""
        names = self.execute_request("select distinct station_name from station")
        return [x[0] for x in names]

    @property
    def ALL_STATION_CODES(self):
        """List of all available station codes in database

        Note
        ----
        Not tested whether the order is the same as the order in
        :attr:`STATION_NAMES`, i.e. the lists should not be linked to each
        other
        """
        names = self.execute_request("select distinct station_code from station")
        return [x[0] for x in names]

    @property
    def ALL_STATISTICS_PARAMS(self):
        """List of all statistical parameters available

        For more info see `here <https://ebas-submit.nilu.no/Submit-Data/
        Data-Reporting/Comments/Generic-metadata-comments/Statistics>`__
        """
        names = self.execute_request("select distinct statistics from variable")
        return [x[0] for x in names]

    @property
    def ALL_VARIABLES(self):
        """List of all variables available"""
        names = self.execute_request("select distinct comp_name from variable")
        return [x[0] for x in names]

    @property
    def ALL_MATRICES(self):
        """List of all matrix values available"""
        names = self.execute_request("select distinct matrix from variable")
        return [x[0] for x in names]

    @property
    def ALL_INSTRUMENTS(self):
        """List of all variables available"""
        names = self.execute_request("select distinct instr_type from variable")
        return [x[0] for x in names]

    def get_table_names(self):
        """Get all table names in SQLite database file"""
        return [
            x[0]
            for x in self.execute_request("SELECT name FROM sqlite_master WHERE type='table';")
        ]

    def get_table_columns(self, table_name):
        """Get all columns of a table in SQLite database file"""
        req = f"select * from {table_name} where 1=0;"
        with sqlite3.connect(self.database) as con:
            cur = con.cursor()
            cur.execute(req)
            return [f[0] for f in cur.description]

    def execute_request(self, request, file_request=False):
        """Connect to database and retrieve data for input request

        Parameters
        ----------
        request : :obj:`EbasSQLRequest` or :obj:`str`
            request specifications

        Returns
        -------
        list
            list of tuples containing the retrieved results. The number of
            items in each tuple corresponds to the number of requested
            parameters (usually one, can be specified in
            :func:`make_query_str` using argument ``what``)

        """
        if isinstance(request, str):
            sql_str = request
        elif isinstance(request, EbasSQLRequest):
            if not file_request:
                sql_str = request.make_query_str()
            else:
                sql_str = request.make_file_query_str()
        else:
            raise ValueError(f"Unsupported request type {type(request)}")

        with sqlite3.connect(self.database) as con:
            cur = con.cursor()
            cur.execute(sql_str)
            return [f for f in cur.fetchall()]

    def get_file_names(self, request):
        """Get all files that match the request specifications

        Parameters
        ----------
        request : :obj:`EbasSQLRequest` or :obj:`str`
            request specifications

        Returns
        -------
        list
            list of file paths that match the request
        """
        return [f[0] for f in self.execute_request(request, file_request=True)]



import logging
import os
from glob import glob

import numpy as np
import pandas as pd
from tqdm import tqdm

from pyaerocom.exceptions import DataRetrievalError
from pyaerocom.io import ReadUngriddedBase
from pyaerocom.stationdata import StationData
from pyaerocom.ungriddeddata import UngriddedData

logger = logging.getLogger(__name__)


class ReadAirNow(ReadUngriddedBase):
    """
    Reading routine for North-American Air Now observations
    """

    # Data type of files
    _FILETYPE = ".dat"

    # File search mask to recursively retrieve list of data files
    _FILEMASK = f"/**/*{_FILETYPE}"

    #: Version log of this class (for caching)
    __version__ = "0.07"

    #: Column delimiter
    FILE_COL_DELIM = "|"

    #: Columns in data files
    FILE_COL_NAMES = [
        "date",
        "time",
        "station_id",
        "station_name",
        "time_zone",
        "variable",
        "unit",
        "value",
        "institute",
    ]

    #: Mapping of columns in station metadata file to pyaerocom standard
    STATION_META_MAP = {
        "aqsid": "station_id",
        "name": "station_name",
        "lat": "latitude",
        "lon": "longitude",
        "elevation": "altitude",
        "city": "city",
        "address": "address",
        "timezone": "timezone",
        "environment": "area_classification",
        "populationclass": "station_classification",
        "modificationdate": "modificationdate",
        "comment": "comment",
    }

    #: conversion functions for metadata dtypes
    STATION_META_DTYPES = {
        "station_id": str,
        "station_name": str,
        "latitude": float,
        "longitude": float,
        "altitude": float,
        "city": str,
        "address": str,
        "timezone": str,
        "area_classification": str,
        "station_classification": str,
        "modificationdate": str,
        "comment": str,
    }

    #: strings to be replaced in original station names
    REPLACE_STATNAME = {"&": "and", "/": " ", ":": " ", ".": " ", "'": ""}

    #: Years in timestamps in the files are are 2-digit (e.g. 20 for 2020)
    BASEYEAR = 2000

    #: Name of dataset (OBS_ID)
    DATA_ID = "AirNow"

    #: List of all datasets supported by this interface
    SUPPORTED_DATASETS = [DATA_ID]

    #: Units found in data files
    UNIT_MAP = {
        "C": "celcius",
        "M/S": "m s-1",
        "MILLIBAR": "mbar",
        "MM": "mm",
        "PERCENT": "%",
        "PPB": "ppb",
        "PPM": "ppm",
        "UG/M3": "ug m-3",
        "WATTS/M2": "W m-2",
    }

    #: Variable names in data files
    VAR_MAP = {
        "concbc": "BC",
        "concpm10": "PM10",
        "concpm25": "PM2.5",
        "vmrco": "CO",
        "vmrnh3": "NH3",
        "vmrno": "NO",
        "vmrno2": "NO2",
        "vmrnox": "NOX",
        "vmrnoy": "NOY",
        "vmro3": "OZONE",
        "vmrso2": "SO2",
    }

    #: List of variables that are provided
    PROVIDES_VARIABLES = list(VAR_MAP)

    #: Default variables
    DEFAULT_VARS = PROVIDES_VARIABLES

    #: Frequncy of measurements
    TS_TYPE = "hourly"

    #: file containing station metadata
    STAT_METADATA_FILENAME = "allStations_20191224.csv"

    def __init__(self, data_id=None, data_dir=None):
        super().__init__(data_id=data_id, data_dir=data_dir)
        self.make_datetime64_array = np.vectorize(self._date_time_str_to_datetime64)
        self._station_metadata = None

    @property
    def station_metadata(self):
        """Dictionary containing global metadata for each site"""
        if self._station_metadata is None:
            self._init_station_metadata()
        return self._station_metadata

    def _date_time_str_to_datetime64(self, date, time):
        """
        Convert date and time string into datetime64 object

        Parameters
        ----------
        date : str
            date string as mm/dd/yy as in data files
        time : str
            time of the day as HH:MM

        Returns
        -------
        datetime64[s]
        """
        mm, dd, yy = date.split("/")
        HH, MM = time.split(":")
        yr = str(self.BASEYEAR + int(yy))
        # returns as datetime64[s]
        return np.datetime64(f"{yr}-{mm}-{dd}T{HH}:{MM}:00")

    def _datetime64_from_filename(self, filepath):
        """
        Get timestamp from filename

        Note
        ----
        This is not in use at the moment and the timestamps in the file may
        differ within 1h around what the filename suggests. So be careful if
        you want to use this method.

        Parameters
        ----------
        filepath : str
            path of file

        Returns
        -------
        datetime64[s]
        """
        fn = os.path.basename(filepath).split(self._FILETYPE)[0]
        assert len(fn) == 10
        tstr = f"{fn[:4]}-{fn[4:6]}-{fn[6:8]}T{fn[8:10]}:00:00"
        return np.datetime64(tstr)

    def _read_metadata_file(self):
        """
        Read station metadatafile

        Returns
        -------
        cfg : pandas.DataFrame
            metadata dataframe

        """
        fn = os.path.join(self.data_dir, self.STAT_METADATA_FILENAME)
        cfg = pd.read_csv(fn, sep=",", converters={"aqsid": lambda x: str(x)})
        return cfg

    def _correct_station_name(self, station_name):
        """
        Remove unwanted chars from original station names

        Parameters
        ----------
        station_name : str
            original station name

        Returns
        -------
        str
            station name cleaned of chars defined in :attr:`REPLACE_STATNAME`

        """
        for search, replace in self.REPLACE_STATNAME.items():
            station_name = station_name.replace(search, replace)
        return station_name

    def _init_station_metadata(self):
        """
        Initiate metadata for all stations

        Returns
        -------
        dict
            dictionary with metadata dictionaries for all stations

        """
        cfg = self._read_metadata_file()
        meta_map = self.STATION_META_MAP

        cols = list(cfg.columns.values)
        col_idx = {}
        for from_meta, to_meta in meta_map.items():
            col_idx[to_meta] = cols.index(from_meta)

        arr = cfg.values
        dtypes = self.STATION_META_DTYPES
        stats = {}
        for row in arr:
            stat = {}
            for meta_key, col_num in col_idx.items():
                stat[meta_key] = dtypes[meta_key](row[col_num])
            sid = stat["station_id"]

            stat["station_name"] = self._correct_station_name(stat["station_name"])
            stat["data_id"] = self.data_id
            stat["ts_type"] = self.TS_TYPE
            stats[sid] = stat
        self._station_metadata = stats
        return stats

    def get_file_list(self):
        """
        Retrieve list of data files

        Returns
        -------
        list
        """
        basepath = self.data_dir
        pattern = f"{basepath}{self._FILEMASK}"
        files = sorted(glob(pattern))
        return files

    def _read_file(self, file):
        """
        Read one datafile using :func:`pandas.read_csv`

        Parameters
        ----------
        file : str
            file path

        Returns
        -------
        df : pandas.DataFrame
            DataFrame containing the file data

        """
        df = pd.read_csv(file, sep=self.FILE_COL_DELIM, names=self.FILE_COL_NAMES)
        return df

    def _read_files(self, files, vars_to_retrieve):
        """
        Read input variables from list of files

        Parameters
        ----------
        files : list
            list of data files
        vars_to_retrieve : list
            list of variables to retrieve

        Raises
        ------
        NotImplementedError
            if several timezones are assigned to the same station
        AttributeError
            if data unit is unkown

        Returns
        -------
        stats : list
            list of StationData objects

        """
        logger.info("Read AirNow data file(s)")
        # initialize empty dataframe
        varcol = self.FILE_COL_NAMES.index("variable")
        arrs = []
        for i in tqdm(range(len(files))):
            fp = files[i]
            filedata = self._read_file(fp)
            arr = filedata.values

            for i, var in enumerate(vars_to_retrieve):
                cond = arr[:, varcol] == self.VAR_MAP[var]
                if i == 0:
                    mask = cond
                else:
                    mask = np.logical_or(mask, cond)
            matches = mask.sum()
            if matches:
                vardata = arr[mask]
                arrs.append(vardata)
        if len(arrs) == 0:
            raise DataRetrievalError("None of the input variables could be found in input list")
        return self._filedata_to_statlist(arrs, vars_to_retrieve)

    def _filedata_to_statlist(self, arrs, vars_to_retrieve):
        """
        Convert loaded filedata into list of StationData objects

        Parameters
        ----------
        arrs : list
            list of numpy arrays extracted from each file
            (see :func:`_read_files`).
        vars_to_retrieve : list
            list of variables to be retrieved from input data.

        Returns
        -------
        stats : list
            list of :class:`StationData` objects, one for each var and station.

        """
        data = np.concatenate(arrs)

        logger.info("Converting filedata to list os StationData")
        stat_meta = self.station_metadata
        stat_ids = list(stat_meta)
        varcol = self.FILE_COL_NAMES.index("variable")
        statcol = self.FILE_COL_NAMES.index("station_id")
        tzonecol = self.FILE_COL_NAMES.index("time_zone")
        unitcol = self.FILE_COL_NAMES.index("unit")
        valcol = self.FILE_COL_NAMES.index("value")

        dtime = self.make_datetime64_array(data[:, 0], data[:, 1])
        stats = []
        for var in vars_to_retrieve:

            # extract only variable data (should speed things up)
            var_in_file = self.VAR_MAP[var]
            mask = data[:, varcol] == var_in_file
            subset = data[mask]
            dtime_subset = dtime[mask]
            statlist = np.unique(subset[:, statcol])
            for stat_id in tqdm(statlist, desc=var):
                if not stat_id in stat_ids:
                    continue
                statmask = subset[:, statcol] == stat_id
                if statmask.sum() == 0:
                    continue
                statdata = subset[statmask]
                timestamps = dtime_subset[statmask]
                # timezone offsets
                toffs = statdata[:, tzonecol].astype(int)
                stat = StationData(**stat_meta[stat_id])

                vals = statdata[:, valcol]
                units = np.unique(statdata[:, unitcol])
                # errors that did not occur in v0 but that may occur
                assert len(units) == 1
                assert units[0] in self.UNIT_MAP
                toffs = toffs.astype("timedelta64[h]")
                timestamps += toffs
                stat["dtime"] = timestamps
                stat["timezone"] = "UTC"
                stat[var] = vals
                unit = self.UNIT_MAP[units[0]]
                stat["var_info"][var] = dict(units=unit)
                stats.append(stat)
        return stats

    def read_file(self):
        """
        This method is not implemented (but needs to be declared for template)

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError("Not needed for these data since the format is unsuitable...")

    def read(self, vars_to_retrieve=None, first_file=None, last_file=None):
        """
        Read variable data

        Parameters
        ----------
        vars_to_retrieve : str or list, optional
            List of variables to be retrieved. The default is None.
        first_file : int, optional
            Index of first file to be read. The default is None, in which case
            index 0 in file list is used.
        last_file : int, optional
            Index of last file to be read. The default is None, in which case
            last index in file list is used.

        Returns
        -------
        data : UngriddedData
            loaded data object.

        """
        if vars_to_retrieve is None:
            vars_to_retrieve = self.DEFAULT_VARS
        elif isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]
        self._init_station_metadata()
        files = self.get_file_list()
        if first_file is None:
            first_file = 0
        if last_file is None:
            last_file = len(files)
        files = files[first_file:last_file]

        stats = self._read_files(files, vars_to_retrieve)

        data = UngriddedData.from_station_data(
            stats, add_meta_keys=["timezone", "area_classification", "station_classification"]
        )

        return data



"""
Module containing helper functions related to iris I/O methods. These contain
reading of Cubes, and some methods to perform quality checks of the data, e.g.

1. checking and correction of time definition
2. number and length of dimension coordinates must match data array
3. Longitude definition from -180 to 180 (corrected if defined on 0 -> 360 intervall)
"""

import logging
from datetime import datetime

import cf_units
import iris
import iris.coords
import iris.util

try:
    # as of iris version 3
    from iris.util import equalise_attributes
except ImportError:
    # old iris version installed
    from iris.experimental.equalise_cubes import equalise_attributes

from pathlib import Path
from traceback import format_exc

import numpy as np

from pyaerocom import const
from pyaerocom.exceptions import (
    FileConventionError,
    NetcdfError,
    UnresolvableTimeDefinitionError,
    VariableDefinitionError,
)
from pyaerocom.helpers import cftime_to_datetime64, make_datetimeindex_from_year
from pyaerocom.io.fileconventions import FileConventionRead
from pyaerocom.io.helpers import add_file_to_log
from pyaerocom.tstype import TsType

logger = logging.getLogger(__name__)


def load_cubes_custom(files, var_name=None, file_convention=None, perform_fmt_checks=True):
    """Load multiple NetCDF files into CubeList

    Note
    ----
    This function does not apply any concatenation or merging of the variable
    data in the individual files, it only loads the files into individual
    instances of :class:`iris.cube.Cube`, which can be accessed via the
    returned list.

    Parameters
    ----------
    files : list
        list of netcdf file paths
    var_name : str
        name of variable to be imported from input files.
    file_convention : :obj:`FileConventionRead`, optional
        Aerocom file convention. If provided, then the data content (e.g.
        dimension definitions) is tested against definition in file name
    perform_fmt_checks : bool
        if True, additional quality checks (and corrections) are (attempted to
        be) performed.

    Returns
    -------
    list
        loaded cube instances.
    list
        list containing all files from which the input variable could be
        successfully loaded.
    """
    cubes = []
    loaded_files = []
    for i, _file in enumerate(files):
        try:
            cube = load_cube_custom(
                file=_file,
                var_name=var_name,
                file_convention=file_convention,
                perform_fmt_checks=perform_fmt_checks,
            )
            cubes.append(cube)
            loaded_files.append(_file)
        except Exception:
            msg = f"Failed to load {_file}. Reason: {format_exc()}"
            logger.warning(msg)

            if const.WRITE_FILEIO_ERR_LOG:
                add_file_to_log(_file, msg)
    return (cubes, loaded_files)


def load_cube_custom(file, var_name=None, file_convention=None, perform_fmt_checks=None):
    """Load netcdf file as iris.Cube

    Parameters
    ----------
    file : str
        netcdf file
    var_name : str
        name of variable to read
    quality_check : bool
        if True, then a quality check of data is performed against the
        information provided in the filename
    file_convention : :obj:`FileConventionRead`, optional
        Aerocom file convention. If provided, then the data content (e.g.
        dimension definitions) is tested against definition in file name
    perform_fmt_checks : bool
        if True, additional quality checks (and corrections) are (attempted to
        be) performed.

    Returns
    -------
    iris.cube.Cube
        loaded data as Cube
    """
    if isinstance(file, Path):
        file = str(file)  # iris load does not like PosixPath
    if perform_fmt_checks is None:
        perform_fmt_checks = const.GRID_IO.PERFORM_FMT_CHECKS
    cube_list = iris.load(file)
    cube = None
    if var_name is None:
        if not len(cube_list) == 1:
            vars_avail = [c.var_name for c in cube_list]
            raise NetcdfError(
                f"Could not load single cube from {file}. Please "
                f"specify var_name. Input file contains the "
                f"following variables: {vars_avail}"
            )
        cube = cube_list[0]
        var_name = cube.var_name
    else:
        for c in cube_list:
            if c.var_name == var_name:
                cube = c
                break
    if cube is None:
        raise NetcdfError(f"Variable {var_name} not available in file {file}")
    if perform_fmt_checks:
        cube = _cube_quality_check(cube, file, file_convention)
    return cube


def _cube_quality_check(cube, file, file_convention=None):
    """Perform quality check of loaded cube data

    This includes the following checks (not all of them may be applicable):

        - Make sure dimensionless variables have unit 1 (and not empty string)

    """
    coords = get_coord_names_cube(cube)
    try:
        cube = _check_cube_unitless(cube)
    except VariableDefinitionError:
        pass

    grid_io = const.GRID_IO
    if grid_io.CHECK_TIME_FILENAME:
        try:
            cube = _check_correct_time_dim(cube, file, file_convention)
        except FileConventionError:
            logger.warning(
                "WARNING: failed to check / validate "
                "time dim. using information in "
                "filename. Reason: invalid file name "
                "convention"
            )

    if grid_io.CHECK_DIM_COORDS:
        cube = check_dim_coords_cube(cube)

    if "time" in coords and grid_io.DEL_TIME_BOUNDS:
        cube.coord("time").bounds = None

    if "longitude" in coords and grid_io.SHIFT_LONS:
        cube = check_and_regrid_lons_cube(cube)
    return cube


def check_and_regrid_lons_cube(cube):
    """Checks and corrects for if longitudes of :attr:`grid` are 0 -> 360

    Note
    ----
    This method checks if the maximum of the current longitudes array
    exceeds 180. Thus, it is not recommended to use this function after
    subsetting a cube, rather, it should be checked directly when the
    file is loaded (cf. :func:`load_input`)

    Parameters
    ----------
    cube : iris.cube.Cube
        gridded data loaded as iris.Cube

    Returns
    -------
    bool
        True, if longitudes were on 0 -> 360 and have been rolled, else
        False
    """
    if cube.coord("longitude").points.max() > 180:
        logger.info(
            "Rearranging longitude dimension from 0 -> 360 definition to -180 -> 180 definition"
        )
        cube = cube.intersection(longitude=(-180, 180))
    return cube


def check_dim_coord_names_cube(cube):

    from pyaerocom import const

    coords = dict(
        lon=const.COORDINFO["lon"], lat=const.COORDINFO["lat"], time=const.COORDINFO["time"]
    )

    for coord in cube.dim_coords:
        cv, cs, cn = coord.var_name, coord.standard_name, coord.long_name
        c = None
        if cv in coords:
            c = coords[cv]
        elif cs in coords:
            c = coords[cs]
        elif cn in coords:
            c = coords[cn]
        if c is not None:
            var_name = c.var_name
            std_name = c.standard_name
            lng_name = c.long_name
            if not coord.var_name == var_name:
                logger.warning(
                    f"Invalid var_name {coord.standard_name} for "
                    f"coord {coord.var_name} in cube. Overwriting with {std_name}"
                )
                coord.var_name = var_name
            if not coord.standard_name == std_name:
                logger.warning(
                    f"Invalid standard_name {coord.standard_name} for "
                    f"coord {coord.var_name} in cube. Overwriting with {std_name}"
                )
                coord.standard_name = std_name
            if not coord.long_name == lng_name:
                logger.warning(
                    f"Invalid long_name {coord.long_name} for "
                    f"coord {coord.var_name} in cube. Overwriting with {lng_name}"
                )
                coord.long_name = lng_name
    return cube


def check_dim_coords_cube(cube):
    """Checks, and if necessary and applicable, updates coords names in Cube

    Parameters
    ----------
    cube : iris.cube.Cube
        input cube

    Returns
    -------
    iris.cube.Cube
        updated or unchanged cube
    """
    cube = check_dim_coord_names_cube(cube)
    return cube


def _check_cube_unitless(cube):
    """Make sure unit in Cube is 1 if variable is dimensionless"""
    var = cube.var_name
    if not var in const.VARS:
        raise VariableDefinitionError(f"No such pyaerocom default variable: {cube.var_name}")

    unit = cf_units.Unit(cube.units)
    if str(const.VARS[var].units) == "1" and unit.is_unknown():
        cube.units = cf_units.Unit("1")
    return cube


def _get_info_from_filename(file, file_convention=None):
    """Load meta-information from filename

    Parameters
    ----------
    """
    if file_convention is None:
        file_convention = FileConventionRead(from_file=file)
    return file_convention.get_info_from_file(file)


def _check_correct_time_dim(cube, file, file_convention=None):
    """Check if time dimension in input Cube is correct

    Note
    -----
    Needs information about time dimension encoded in filename, since the
    check is done against what is specified in the filename. E.g. AeroCom
    format
    Parameters
    ----------
    cube : iris.cube.Cube
        loaded Cube instance, for which time dimension is supposed to be
        checked
    file : str
        path to file from which the Cube was imported
    file_convention : FileConventionRead
        file naming convention specifying how time dimension information is
        encoded in the filenames.

    """
    finfo = _get_info_from_filename(file, file_convention)

    ts_type = TsType(finfo["ts_type"])
    year = finfo["year"]

    if not const.MIN_YEAR <= year <= const.MAX_YEAR:
        raise FileConventionError(f"Invalid year in file: {year}")
    elif year == 9999:
        logger.info(
            "Cannot compare NetCDF time dimension for climatological data "
            "(9999 in filename). Skipping this check."
        )
    else:
        try:
            check_time_coord(cube, ts_type, year)
        except UnresolvableTimeDefinitionError as e:
            raise UnresolvableTimeDefinitionError(repr(e))
        except Exception as e:
            msg = f"Invalid time dimension coordinate in file:\n{file}.\nError: repr({e})\n"
            logger.warning(msg)
            if const.GRID_IO.CORRECT_TIME_FILENAME:
                add_msg = "Attempting to correct time coordinate using information in file name"
                msg += add_msg
                logger.info(add_msg)
                try:
                    cube = correct_time_coord(cube, ts_type=finfo["ts_type"], year=finfo["year"])
                except Exception:
                    add_msg = (
                        f"Unable to correct time dimension using the "
                        f"information provided in the file name. Error:\n"
                        f"{format_exc()}.\n\nThe file will be imported regardless!"
                    )
                    msg += add_msg
                    logger.warning(msg)
            if const.WRITE_FILEIO_ERR_LOG:
                add_file_to_log(file, msg)
    return cube


def _check_leap_year(num, num_per, ts_type):
    if ts_type == "daily" and num + 1 == num_per:
        return True
    elif ts_type == "3hourly" and num + 8 == num_per:
        return True
    elif ts_type == "hourly" and num + 24 == num_per:
        return True
    return False


def check_time_coord(cube, ts_type, year):
    """Method that checks the time coordinate of an iris Cube

    This method checks if the time dimension of a cube is accessible and
    according to the standard (i.e. fully usable). It only checks, and does not
    correct. For the latter, please see :func:`correct_time_coord`.

    Parameters
    ----------
    cube : Cube
        cube containing data
    ts_type : str
        pyaerocom ts_type
    year :
        year of data

    Returns
    -------
    bool
        True, if time dimension is ok, False if not
    """
    if isinstance(ts_type, str):
        ts_type = TsType(ts_type)
    if not "time" in get_coord_names_cube(cube):
        raise AttributeError("Cube does not contain time dimension")
    tdim = cube.coord("time")

    if not isinstance(tdim, iris.coords.DimCoord):
        raise AttributeError("Time is not a DimCoord instance")
    try:
        cftime_to_datetime64(0, cfunit=tdim.units)
    except Exception:
        raise ValueError("Could not convert time unit string")

    freq = ts_type.to_pandas_freq()

    tidx = make_datetimeindex_from_year(freq, year)

    num_per = len(tidx)
    num = len(tdim.points)

    if not num == num_per:
        if tidx[0].is_leap_year:
            if not _check_leap_year(num, num_per, ts_type):
                raise UnresolvableTimeDefinitionError(
                    f"Expected {len(tidx)} timestamps but data has {num}"
                )

        else:
            raise UnresolvableTimeDefinitionError(
                f"Expected {len(tidx)} timestamps but data has {num}"
            )

    # ToDo: check why MS is not working for period conversion
    if freq == "MS":
        freq = "M"
    # convert first and last timestamps of index array into periods
    # (e.g. January and December for monthly data)
    per0 = tidx[0].to_period(freq)
    per1 = tidx[-1].to_period(freq)

    # first and last timestamp in data
    t0, t1 = cftime_to_datetime64([tdim.points[0], tdim.points[-1]], cfunit=tdim.units)

    if not per0.start_time <= t0 <= per0.end_time:
        raise ValueError(f"First timestamp of data {t0} does not lie in first period: {per0}")
    elif not per1.start_time <= t1 <= per1.end_time:
        raise ValueError(f"Last timestamp of data {t1} does not lie in end period: {per1}")


def get_dim_names_cube(cube):
    return [c.name() for c in cube.dim_coords]


def get_coord_names_cube(cube):
    return [c.name() for c in cube.coords()]


def _get_time_index_cube(cube):
    """
    Get array index of time dimension for input Cube

    Parameters
    ----------
    cube : iris.cube.Cube
        data cube.

    Raises
    ------
    IndexError
        if index cannot be retrieved (e.g. data does not contain time
        dimension).

    Returns
    -------
    int

    """
    dim_names = get_dim_names_cube(cube)
    if "time" in dim_names:
        return dim_names.index("time")

    idx_miss = []
    if cube.ndim != len(dim_names):  # one dimension is missing
        for idx in range(len(cube.shape)):
            coords = cube.coords(contains_dimension=idx, dim_coords=True)
            if len(coords) == 0:
                idx_miss.append(idx)
    if len(idx_miss) == 1:
        return idx_miss[0]

    raise IndexError(f"Failed to identify data index of time dimension in cube {repr(cube)}")


def correct_time_coord(cube, ts_type, year):
    """Method that corrects the time coordinate of an iris Cube

    Parameters
    ----------
    cube : Cube
        cube containing data
    ts_type : TsType or str
        temporal resolution of data (e.g. "hourly", "daily"). This information
        is e.g. encoded in the filename of a NetCDF file and may be
        accessed using :class:`pyaerocom.io.FileConventionRead`
    year : int
        integer specifying start year, e.g. 2017

    Returns
    -------
    Cube
        the same instance of the input cube with corrected time dimension axis

    """
    tindex_cube = _get_time_index_cube(cube)
    coords = get_coord_names_cube(cube)

    if isinstance(ts_type, str):
        ts_type = TsType(ts_type)

    tres_str = ts_type.cf_base_unit
    conv = ts_type.datetime64_str
    tunit_str = f"{tres_str} since {year}-01-01 00:00:00"
    num = cube.shape[tindex_cube]

    tunit = cf_units.Unit(tunit_str, calendar=cf_units.CALENDAR_STANDARD)
    tres_np = ts_type.timedelta64_str  # TSTR_TO_NP_TD[ts_type]
    base = np.datetime64(f"{year}-01-01 00:00:00").astype(conv)
    times = base + np.arange(0, num, 1).astype(tres_np)

    # see this thread https://github.com/matplotlib/matplotlib/issues/2259/
    times_dt = times.astype("datetime64[s]").astype(datetime)

    time_nums = [tunit.date2num(t) for t in times_dt]

    pd_freq = ts_type.to_pandas_freq()
    num_expected = len(make_datetimeindex_from_year(pd_freq, year))
    num_inferred = len(time_nums)
    if not num_inferred == num_expected:
        raise UnresolvableTimeDefinitionError(
            f"expected {num_expected} timestamps for {year} and "
            f"freq {ts_type} but got {num_inferred}"
        )
    tcoord = iris.coords.DimCoord(time_nums, standard_name="time", units=tunit)
    if "time" in coords:
        cube.remove_coord("time")

    cube.add_dim_coord(tcoord, tindex_cube)
    cube.attributes["timedim-corrected"] = True
    return cube


def _check_correct_dtypes_timedim_cube_list(cubes):

    try:
        dtypes = np.unique([cube.coord("time").points.dtype for cube in cubes])
    except iris.exceptions.CoordinateNotFoundError:
        return False
    corrected = False
    if len(dtypes) > 1:
        corrected = True
        for cube in cubes:
            new = cube.coord("time").points.astype(float)
            cube.coord("time").points = new
    return corrected


def concatenate_iris_cubes(cubes, error_on_mismatch=True):
    """Concatenate list of :class:`iris.Cube` instances cubes into single Cube

    Helper method for concatenating list of cubes

    This method is not supposed to be called directly but rather
    :func:`concatenate_cubes` (which ALWAYS returns instance of
    :class:`Cube` or raises Exception) or :func:`concatenate_possible_cubes`
    (which ALWAYS returns instance of :class:`CubeList` or raises Exception)

    Parameters
    ----------
    cubes : CubeList or list(Cubes)
        list of individual cubes
    error_on_mismatch
        boolean specifying whether an Exception is supposed to be raised
        or not

    Returns
    -------
    :obj:`Cube`
        result of concatenation

    Raises
    ------
    iris.exceptions.ConcatenateError
        if ``error_on_mismatch=True`` and input cubes could not all concatenated
        into a single instance of :class:`iris.Cube` class.

    """
    cubes = iris.cube.CubeList(cubes)
    var_name = cubes[0].var_name
    if const.GRID_IO.EQUALISE_METADATA:
        meta_init = cubes[0].metadata
        if not all([x.metadata == meta_init for x in cubes]):
            logger.warning(
                f"{var_name} cubes to be concatenated have different meta data settings. "
                f"These will be unified using the metadata dictionary of the first cube "
                f"(otherwise the method concatenate of the iris package won't work)"
            )
            for cube in cubes:
                cube.metadata = meta_init

    # now put the CubeList together and form one cube
    # 1st equalise the cubes (remove non common attributes)
    equalise_attributes(cubes)
    # unify time units
    iris.util.unify_time_units(cubes)
    # now concatenate the cube list to one cube

    try:
        cubes_concat = iris.cube.CubeList.concatenate_cube(cubes, error_on_mismatch)
    except Exception as e:
        if _check_correct_dtypes_timedim_cube_list(cubes):
            cubes_concat = iris.cube.CubeList.concatenate_cube(cubes, error_on_mismatch)
        else:
            raise

    return cubes_concat



import os

import numpy as np
import pandas as pd

from pyaerocom import const
from pyaerocom.aux_var_helpers import calc_od550aer, calc_od550gt1aer, calc_od550lt1aer
from pyaerocom.io.readaeronetbase import ReadAeronetBase
from pyaerocom.stationdata import StationData


class ReadAeronetSdaV3(ReadAeronetBase):
    """Interface for reading Aeronet Sun SDA V3 Level 1.5 and 2.0 data

    .. seealso::

        Base classes :class:`ReadAeronetBase` and :class:`ReadUngriddedBase`

    """

    #: Mask for identifying datafiles
    _FILEMASK = "*.lev30"

    #: version log of this class (for caching)
    __version__ = "0.08_" + ReadAeronetBase.__baseversion__

    #: Name of dataset (OBS_ID)
    DATA_ID = const.AERONET_SUN_V3L2_SDA_DAILY_NAME

    #: List of all datasets supported by this interface
    SUPPORTED_DATASETS = [
        const.AERONET_SUN_V3L15_SDA_DAILY_NAME,
        const.AERONET_SUN_V3L2_SDA_DAILY_NAME,
    ]

    #: dictionary assigning temporal resolution flags for supported datasets
    #: that are provided in a defined temporal resolution
    TS_TYPES = {
        const.AERONET_SUN_V3L15_SDA_DAILY_NAME: "daily",
        const.AERONET_SUN_V3L2_SDA_DAILY_NAME: "daily",
    }

    #: default variables for read method
    DEFAULT_VARS = ["od550aer", "od550gt1aer", "od550lt1aer", "od550dust"]

    #: value corresponding to invalid measurement
    NAN_VAL = -999.0

    #: dictionary specifying the file column names (values) for each Aerocom
    #: variable (keys)
    VAR_NAMES_FILE = {}
    VAR_NAMES_FILE["od500gt1aer"] = "Coarse_Mode_AOD_500nm[tau_c]"
    VAR_NAMES_FILE["od500lt1aer"] = "Fine_Mode_AOD_500nm[tau_f]"
    VAR_NAMES_FILE["od500aer"] = "Total_AOD_500nm[tau_a]"
    VAR_NAMES_FILE["ang4487aer"] = "Angstrom_Exponent(AE)-Total_500nm[alpha]"
    VAR_NAMES_FILE["od500dust"] = "Coarse_Mode_AOD_500nm[tau_c]"

    #: dictionary specifying the file column names (values) for each
    #: metadata key (cf. attributes of :class:`StationData`, e.g.
    #: 'station_name', 'longitude', 'latitude', 'altitude')
    META_NAMES_FILE = {}
    META_NAMES_FILE["data_quality_level"] = "Data_Quality_Level"
    META_NAMES_FILE["instrument_number"] = "AERONET_Instrument_Number"
    META_NAMES_FILE["station_name"] = "AERONET_Site"
    META_NAMES_FILE["latitude"] = "Site_Latitude(Degrees)"
    META_NAMES_FILE["longitude"] = "Site_Longitude(Degrees)"
    META_NAMES_FILE["altitude"] = "Site_Elevation(m)"
    META_NAMES_FILE["date"] = "Date_(dd:mm:yyyy)"
    META_NAMES_FILE["time"] = "Time_(hh:mm:ss)"
    META_NAMES_FILE["day_of_year"] = "Day_of_Year"

    #: dictionary containing information about additionally required variables
    #: for each auxiliary variable (i.e. each variable that is not provided
    #: by the original data but computed on import)
    AUX_REQUIRES = {
        "od550aer": ["od500aer", "ang4487aer"],
        "od550gt1aer": ["od500gt1aer", "ang4487aer"],
        "od550dust": ["od500gt1aer", "ang4487aer"],
        "od550lt1aer": ["od500lt1aer", "ang4487aer"],
    }

    #: Functions that are used to compute additional variables (i.e. one
    #: for each variable defined in AUX_REQUIRES)
    AUX_FUNS = {
        "od550aer": calc_od550aer,
        "od550gt1aer": calc_od550gt1aer,
        "od550dust": calc_od550gt1aer,
        "od550lt1aer": calc_od550lt1aer,
    }

    #: List of variables that are provided by this dataset (will be extended
    #: by auxiliary variables on class init, for details see __init__ method of
    #: base class ReadUngriddedBase)
    PROVIDES_VARIABLES = list(VAR_NAMES_FILE)

    def read_file(self, filename, vars_to_retrieve=None, vars_as_series=False):
        """Read Aeronet SDA V3 file and return it in a dictionary

        Parameters
        ----------
        filename : str
            absolute path to filename to read
        vars_to_retrieve : :obj:`list`, optional
            list of str with variable names to read. If None, use
            :attr:`DEFAULT_VARS`
        vars_as_series : bool
            if True, the data columns of all variables in the result dictionary
            are converted into pandas Series objects

        Returns
        -------
        StationData
            dict-like object containing results
        """
        if vars_to_retrieve is None:
            vars_to_retrieve = self.DEFAULT_VARS
        # implemented in base class
        vars_to_read, vars_to_compute = self.check_vars_to_retrieve(vars_to_retrieve)

        # create empty data object (is dictionary with extended functionality)
        data_out = StationData()
        data_out.data_id = self.data_id
        # create empty arrays for meta information
        for item in self.META_NAMES_FILE:
            data_out[item] = []

        # create empty arrays for all variables that are supposed to be read
        # from file
        for var in vars_to_read:
            data_out[var] = []

        # Iterate over the lines of the file
        self.logger.info(f"Reading file {filename}")

        with open(filename) as in_file:
            # skip first 4 lines
            in_file.readline()
            in_file.readline()
            in_file.readline()
            in_file.readline()

            # PI line
            dummy_arr = in_file.readline().strip().split(";")
            data_out["PI"] = dummy_arr[0].split("=")[1]
            data_out["PI_email"] = dummy_arr[1].split("=")[1]
            data_out["ts_type"] = self.TS_TYPE

            # skip this line
            in_file.readline()

            # put together a dict with the header string as key and the index number as value so that we can access
            # the index number via the header string
            col_index_str = in_file.readline()
            if col_index_str != self._last_col_index_str:
                self.logger.info("Header has changed, reloading col_index map")
                self._update_col_index(col_index_str)
            col_index = self.col_index

            # dependent on the station, some of the required input variables
            # may not be provided in the data file. These will be ignored
            # in the following list that iterates over all data rows and will
            # be filled below, with vectors containing NaNs after the file
            # reading loop
            vars_available = {}
            for var in vars_to_read:
                if var in col_index:
                    vars_available[var] = col_index[var]
                else:
                    self.logger.warning(
                        f"Variable {var} not available in file {os.path.basename(filename)}"
                    )

            for line in in_file:
                # process line
                dummy_arr = line.split(self.COL_DELIM)

                # copy the meta data (array of type string)
                for var in self.META_NAMES_FILE:
                    val = dummy_arr[col_index[var]]
                    try:
                        # e.g. lon, lat, altitude
                        val = float(val)
                    except Exception:
                        pass
                    data_out[var].append(val)

                # This uses the numpy datestring64 functions that e.g. also support Months as a time step for timedelta
                # Build a proper ISO 8601 UTC date string

                day, month, year = dummy_arr[col_index["date"]].split(":")
                datestring = "-".join([year, month, day])
                datestring = "T".join([datestring, dummy_arr[col_index["time"]]])
                # NOTE JGLISS: parsing timezone offset was removed on 22/2/19
                # since it is deprecated in recent numpy versions, for details
                # see https://www.numpy.org/devdocs/reference/arrays.datetime.html#changes-with-numpy-1-11
                # datestring = '+'.join([datestring, '00:00'])

                data_out["dtime"].append(np.datetime64(datestring))

                # copy the data fields
                for var, idx in vars_available.items():
                    val = np.float_(dummy_arr[idx])
                    if val == self.NAN_VAL:
                        val = np.nan
                    data_out[var].append(val)

        # convert all lists to numpy arrays
        data_out["dtime"] = np.asarray(data_out["dtime"])

        for item in self.META_NAMES_FILE:
            data_out[item] = np.asarray(data_out[item])

        for var in vars_to_read:
            if var in vars_available:
                array = np.asarray(data_out[var])
            else:
                array = np.zeros(len(data_out["dtime"])) * np.nan
            data_out[var] = array

        # compute additional variables (if applicable)
        data_out = self.compute_additional_vars(data_out, vars_to_compute)

        # convert data vectors to pandas.Series (if applicable)
        if vars_as_series:
            for var in vars_to_read + vars_to_compute:
                if var in vars_to_retrieve:
                    data_out[var] = pd.Series(data_out[var], index=data_out["dtime"])
                else:
                    del data_out[var]

        return data_out



import gzip
import logging
import os
import pathlib
import shutil
import tempfile

import numpy as np
import pandas as pd

from pyaerocom import const
from pyaerocom.aux_var_helpers import calc_ang4487aer, calc_od550aer, calc_od550lt1ang
from pyaerocom.exceptions import AeronetReadError
from pyaerocom.io.readaeronetbase import ReadAeronetBase
from pyaerocom.stationdata import StationData

logger = logging.getLogger(__name__)


class ReadAeronetSunV3(ReadAeronetBase):
    """Interface for reading Aeronet direct sun version 3 Level 1.5 and 2.0 data

    .. seealso::

        Base classes :class:`ReadAeronetBase` and :class:`ReadUngriddedBase`

    """

    #: Mask for identifying datafiles
    _FILEMASK = "*.lev*"

    #: version log of this class (for caching)
    __version__ = "0.10_" + ReadAeronetBase.__baseversion__

    #: Name of dataset (OBS_ID)
    DATA_ID = const.AERONET_SUN_V3L2_AOD_DAILY_NAME

    #: List of all datasets supported by this interface
    SUPPORTED_DATASETS = [
        const.AERONET_SUN_V3L15_AOD_DAILY_NAME,
        const.AERONET_SUN_V3L15_AOD_ALL_POINTS_NAME,
        const.AERONET_SUN_V3L2_AOD_DAILY_NAME,
        const.AERONET_SUN_V3L2_AOD_ALL_POINTS_NAME,
    ]

    #: dictionary assigning temporal resolution flags for supported datasets
    #: that are provided in a defined temporal resolution
    TS_TYPES = {
        const.AERONET_SUN_V3L15_AOD_DAILY_NAME: "daily",
        const.AERONET_SUN_V3L2_AOD_DAILY_NAME: "daily",
    }

    #: default variables for read method
    DEFAULT_VARS = ["od550aer", "ang4487aer"]

    #: value corresponding to invalid measurement
    # NAN_VAL = -9999.
    NAN_VAL = -999.0

    #: Mappings for identifying variables in file
    VAR_PATTERNS_FILE = {"AOD_([0-9]*)nm": "od*aer"}

    #: dictionary specifying the file column names (values) for each Aerocom
    #: variable (keys)
    VAR_NAMES_FILE = {}
    VAR_NAMES_FILE["od340aer"] = "AOD_340nm"
    VAR_NAMES_FILE["od440aer"] = "AOD_440nm"
    VAR_NAMES_FILE["od500aer"] = "AOD_500nm"
    # VAR_NAMES_FILE['od865aer'] = 'AOD_865nm'
    VAR_NAMES_FILE["od870aer"] = "AOD_870nm"
    VAR_NAMES_FILE["ang4487aer"] = "440-870_Angstrom_Exponent"

    #: dictionary specifying the file column names (values) for each
    #: metadata key (cf. attributes of :class:`StationData`, e.g.
    #: 'station_name', 'longitude', 'latitude', 'altitude')
    META_NAMES_FILE = {}
    META_NAMES_FILE["data_quality_level"] = "Data_Quality_Level"
    META_NAMES_FILE["instrument_number"] = "AERONET_Instrument_Number"
    META_NAMES_FILE["station_name"] = "AERONET_Site"
    META_NAMES_FILE["latitude"] = "Site_Latitude(Degrees)"
    META_NAMES_FILE["longitude"] = "Site_Longitude(Degrees)"
    META_NAMES_FILE["altitude"] = "Site_Elevation(m)"
    META_NAMES_FILE["date"] = "Date(dd:mm:yyyy)"
    META_NAMES_FILE["time"] = "Time(hh:mm:ss)"
    META_NAMES_FILE["day_of_year"] = "Day_of_Year"

    META_NAMES_FILE_ALT = {"AERONET_Site": ["AERONET_Site_Name"]}
    #: dictionary containing information about additionally required variables
    #: for each auxiliary variable (i.e. each variable that is not provided
    #: by the original data but computed on import)
    AUX_REQUIRES = {
        "ang44&87aer": ["od440aer", "od870aer"],
        "od550aer": ["od440aer", "od500aer", "ang4487aer"],
        "od550lt1ang": ["od440aer", "od500aer", "ang4487aer"],
    }

    #: Functions that are used to compute additional variables (i.e. one
    #: for each variable defined in AUX_REQUIRES)
    AUX_FUNS = {
        "ang44&87aer": calc_ang4487aer,
        "od550aer": calc_od550aer,
        "od550lt1ang": calc_od550lt1ang,
    }

    #: List of variables that are provided by this dataset (will be extended
    #: by auxiliary variables on class init, for details see __init__ method of
    #: base class ReadUngriddedBase)
    PROVIDES_VARIABLES = list(VAR_NAMES_FILE)

    def read_file(self, filename, vars_to_retrieve=None, vars_as_series=False):
        """Read Aeronet Sun V3 level 1.5 or 2 file

        Parameters
        ----------
        filename : str
            absolute path to filename to read
        vars_to_retrieve : :obj:`list`, optional
            list of str with variable names to read. If None, use
            :attr:`DEFAULT_VARS`
        vars_as_series : bool
            if True, the data columns of all variables in the result dictionary
            are converted into pandas Series objects

        Returns
        -------
        StationData
            dict-like object containing results
        """
        if vars_to_retrieve is None:
            vars_to_retrieve = self.DEFAULT_VARS
        # implemented in base class
        vars_to_read, vars_to_compute = self.check_vars_to_retrieve(vars_to_retrieve)

        # create empty data object (is dictionary with extended functionality)
        data_out = StationData()
        data_out.data_id = self.data_id
        # create empty arrays for meta information
        for item in self.META_NAMES_FILE:
            data_out[item] = []

        # Iterate over the lines of the file
        self.logger.info(f"Reading file {filename}")
        # enable alternative reading of .gz files here to save space on the file system
        suffix = pathlib.Path(filename).suffix
        tmp_name = filename
        if suffix == ".gz":
            f_out = tempfile.NamedTemporaryFile(delete=False)
            with gzip.open(filename, "r") as f_in:
                shutil.copyfileobj(f_in, f_out)
            filename = f_out.name
            f_out.close()

        try:
            with open(filename, "rt") as in_file:
                lines = in_file.readlines()
        except UnicodeDecodeError:
            with open(filename, "rt", encoding="ISO-8859-1") as in_file:
                lines = in_file.readlines()
        except OSError:
            # faulty gzip file, but also the gzip class raises some exceptions
            if suffix == ".gz":
                os.remove(f_out.name)
            raise AeronetReadError(f"gzip error in file {tmp_name}")

        _lines_ignored = []

        line_idx = 4
        _lines_ignored.append(lines[0 : line_idx - 1])

        # PI line
        dummy_arr = lines[line_idx].strip().split(";")
        line_idx += 1
        data_out["PI"] = dummy_arr[0].split("=")[1]
        data_out["PI_email"] = dummy_arr[1].split("=")[1]
        data_out["ts_type"] = self.TS_TYPE

        data_type_comment = lines[line_idx]
        line_idx += 1

        _lines_ignored.append(data_type_comment)
        self.logger.debug(f"Data type comment: {data_type_comment}")

        # put together a dict with the header string as key and the index number as value so that we can access
        # the index number via the header string
        col_index_str = lines[line_idx]
        line_idx += 1

        if col_index_str != self._last_col_index_str:
            self.logger.info("Header has changed, reloading col_index map")
            self._update_col_index(col_index_str)
        col_index = self.col_index

        # dependent on the station, some of the required input variables
        # may not be provided in the data file. These will be ignored
        # in the following list that iterates over all data rows and will
        # be filled below, with vectors containing NaNs after the file
        # reading loop
        vars_available = {}
        for var in vars_to_read:
            data_out[var] = []
            if var in col_index:
                vars_available[var] = col_index[var]
            else:
                self.logger.warning(
                    f"Variable {var} not available in file {os.path.basename(filename)}"
                )
        pl = None

        for i, line in enumerate(lines[line_idx:]):
            # process line
            dummy_arr = line.split(self.COL_DELIM)

            if pl is not None and len(dummy_arr) != len(pl):
                self.logger.warning(f"Data line {i} in {filename} is corrupt, skipping...")
                continue
            # copy the meta data (array of type string)
            for var in self.META_NAMES_FILE:
                try:
                    val = dummy_arr[col_index[var]]
                except IndexError as e:
                    self.logger.warning(repr(e))

                try:
                    # e.g. lon, lat, altitude
                    val = float(val)
                except Exception:
                    pass
                data_out[var].append(val)

            # This uses the numpy datestring64 functions that e.g. also
            # support Months as a time step for timedelta
            # Build a proper ISO 8601 UTC date string
            day, month, year = dummy_arr[col_index["date"]].split(":")
            datestring = "-".join([year, month, day])
            datestring = "T".join([datestring, dummy_arr[col_index["time"]]])
            # NOTE JGLISS: parsing timezone offset was removed on 22/2/19
            # since it is deprecated in recent numpy versions, for details
            # see https://www.numpy.org/devdocs/reference/arrays.datetime.html#changes-with-numpy-1-11
            # datestring = '+'.join([datestring, '00:00'])

            data_out["dtime"].append(np.datetime64(datestring))

            for var, idx in vars_available.items():
                val = np.float_(dummy_arr[idx])
                if val == self.NAN_VAL:
                    val = np.nan
                data_out[var].append(val)

            pl = dummy_arr

        # remove the temp file in case the input file was a gz file
        if suffix == ".gz":
            os.remove(f_out.name)

        # convert all lists to numpy arrays
        data_out["dtime"] = np.asarray(data_out["dtime"])

        for item in self.META_NAMES_FILE:
            data_out[item] = np.asarray(data_out[item])

        for var in vars_to_read:
            if var in vars_available:
                array = np.asarray(data_out[var])
            else:
                array = np.zeros(len(data_out["dtime"])) * np.nan
            data_out[var] = array

        # compute additional variables (if applicable)
        data_out = self.compute_additional_vars(data_out, vars_to_compute)

        # convert data vectors to pandas.Series (if applicable)
        if vars_as_series:
            for var in vars_to_read + vars_to_compute:
                if var in vars_to_retrieve:
                    data_out[var] = pd.Series(data_out[var], index=data_out["dtime"])
                else:
                    del data_out[var]
        self.logger.debug(f"The following lines were ignored: {_lines_ignored}")
        return data_out



import logging
import os
from time import time

from pyaerocom import ColocatedData
from pyaerocom._lowlevel_helpers import write_json
from pyaerocom.aeroval._processing_base import ProcessingEngine
from pyaerocom.aeroval.coldatatojson_helpers import (
    _add_entry_json,
    _apply_annual_constraint,
    _init_data_default_frequencies,
    _init_meta_glob,
    _process_heatmap_data,
    _process_map_and_scat,
    _process_regional_timeseries,
    _process_sites,
    _process_sites_weekly_ts,
    _process_statistics_timeseries,
    _write_site_data,
    _write_stationdata_json,
    get_heatmap_filename,
    get_json_mapname,
    init_regions_web,
    update_regions_json,
)
from pyaerocom.exceptions import AeroValConfigError, TemporalResolutionError

logger = logging.getLogger(__name__)


class ColdataToJsonEngine(ProcessingEngine):
    def run(self, files):
        """
        Convert colocated data files to json

        Parameters
        ----------
        files : list
            list of file paths pointing to colocated data objects to be
            processed.

        Returns
        -------
        list
            list of files that have been converted.

        """
        converted = []
        for file in files:
            logger.info(f"Processing: {file}")
            coldata = ColocatedData(file)
            self.process_coldata(coldata)
            converted.append(file)
        return converted

    def process_coldata(self, coldata: ColocatedData):
        """
        Creates all json files for one ColocatedData object

        Parameters
        ----------
        coldata : ColocatedData
            colocated data to be processed.

        Raises
        ------
        NotImplementedError
            DESCRIPTION.
        ValueError
            DESCRIPTION.
        AeroValConfigError
            DESCRIPTION.

        Returns
        -------
        None.

        """
        t00 = time()
        use_weights = self.cfg.statistics_opts.weighted_stats
        # redundant, but cheap and important to be correct
        self.cfg._check_time_config()
        freqs = self.cfg.time_cfg.freqs
        periods = self.cfg.time_cfg.periods
        seasons = self.cfg.time_cfg.get_seasons()
        main_freq = self.cfg.time_cfg.main_freq
        annual_stats_constrained = self.cfg.statistics_opts.annual_stats_constrained

        out_dirs = self.cfg.path_manager.get_json_output_dirs(True)
        regions_json = self.exp_output.regions_file
        regions_how = self.cfg.webdisp_opts.regions_how

        stats_min_num = self.cfg.statistics_opts.MIN_NUM

        vert_code = coldata.get_meta_item("vert_code")
        diurnal_only = coldata.get_meta_item("diurnal_only")

        add_trends = self.cfg.statistics_opts.add_trends
        trends_min_yrs = self.cfg.statistics_opts.trends_min_yrs

        use_fairmode = self.cfg.statistics_opts.use_fairmode

        # ToDo: some of the checks below could be done automatically in
        # EvalSetup, and at an earlier stage
        if vert_code == "ModelLevel":
            raise NotImplementedError("Coming (not so) soon...")

        # this will need to be figured out as soon as there is altitude
        elif "altitude" in coldata.data.dims:
            raise NotImplementedError("Cannot yet handle profile data")

        elif not isinstance(coldata, ColocatedData):
            raise ValueError(f"Need ColocatedData object, got {type(coldata)}")

        elif coldata.has_latlon_dims and regions_how == "country":
            raise NotImplementedError(
                "Cannot yet apply country filtering for 4D colocated data instances"
            )
        elif not main_freq in freqs:
            raise AeroValConfigError(
                f"Scatter plot frequency {main_freq} is not in experiment frequencies: {freqs}"
            )
        if self.cfg.statistics_opts.stats_tseries_base_freq is not None:
            if not self.cfg.statistics_opts.stats_tseries_base_freq in freqs:
                raise AeroValConfigError(
                    f"Base frequency for statistics timeseries needs to be "
                    f"specified in experiment frequencies: {freqs}"
                )
        # init some stuff
        if "var_name_input" in coldata.metadata:
            obs_var = coldata.metadata["var_name_input"][0]
            model_var = coldata.metadata["var_name_input"][1]
        else:
            obs_var = model_var = "UNDEFINED"

        model_name = coldata.model_name
        obs_name = coldata.obs_name

        mcfg = self.cfg.model_cfg.get_entry(model_name)
        var_name_web = mcfg.get_varname_web(model_var, obs_var)

        logger.info(
            f"Computing json files for {model_name} ({model_var}) vs. {obs_name} ({obs_var})"
        )

        meta_glob = _init_meta_glob(
            coldata,
            vert_code=vert_code,
            obs_name=obs_name,
            model_name=model_name,
            var_name_web=var_name_web,
        )

        # get region IDs
        (regborders, regs, regnames) = init_regions_web(coldata, regions_how)

        update_regions_json(regborders, regions_json)

        use_country = True if regions_how == "country" else False

        data = _init_data_default_frequencies(coldata, freqs)

        if annual_stats_constrained:
            data = _apply_annual_constraint(data)

        if not diurnal_only:
            logger.info("Processing statistics timeseries for all regions")
            input_freq = self.cfg.statistics_opts.stats_tseries_base_freq
            try:
                stats_ts = _process_statistics_timeseries(
                    data=data,
                    freq=main_freq,
                    region_ids=regnames,
                    use_weights=use_weights,
                    use_country=use_country,
                    data_freq=input_freq,
                )
            except TemporalResolutionError:
                stats_ts = {}

            ts_file = os.path.join(out_dirs["hm/ts"], "stats_ts.json")
            _add_entry_json(
                ts_file, stats_ts, obs_name, var_name_web, vert_code, model_name, model_var
            )

            logger.info("Processing heatmap data for all regions")
            hm_all = _process_heatmap_data(
                data,
                regnames,
                use_weights,
                use_country,
                meta_glob,
                periods,
                seasons,
                add_trends,
                trends_min_yrs,
            )

            for freq, hm_data in hm_all.items():
                fname = get_heatmap_filename(freq)

                hm_file = os.path.join(out_dirs["hm"], fname)

                _add_entry_json(
                    hm_file, hm_data, obs_name, var_name_web, vert_code, model_name, model_var
                )

            logger.info("Processing regional timeseries for all regions")
            ts_objs_regional = _process_regional_timeseries(data, regnames, regions_how, meta_glob)

            _write_site_data(ts_objs_regional, out_dirs["ts"])
            if coldata.has_latlon_dims:
                for cd in data.values():
                    if cd is not None:
                        cd.data = cd.flatten_latlondim_station_name().data

            logger.info("Processing individual site timeseries data")
            (ts_objs, map_meta, site_indices) = _process_sites(data, regs, regions_how, meta_glob)

            _write_site_data(ts_objs, out_dirs["ts"])

            map_data, scat_data = _process_map_and_scat(
                data,
                map_meta,
                site_indices,
                periods,
                main_freq,
                stats_min_num,
                seasons,
                add_trends,
                trends_min_yrs,
                use_fairmode,
                obs_var,
            )

            map_name = get_json_mapname(obs_name, var_name_web, model_name, model_var, vert_code)

            outfile_map = os.path.join(out_dirs["map"], map_name)
            write_json(map_data, outfile_map, ignore_nan=True)

            outfile_scat = os.path.join(out_dirs["scat"], map_name)
            write_json(scat_data, outfile_scat, ignore_nan=True)

        if coldata.ts_type == "hourly":
            logger.info("Processing diurnal profiles")
            (ts_objs_weekly, ts_objs_weekly_reg) = _process_sites_weekly_ts(
                coldata, regions_how, regnames, meta_glob
            )
            outdir = os.path.join(out_dirs["ts/diurnal"])
            for ts_data_weekly in ts_objs_weekly:
                # writes json file
                _write_stationdata_json(ts_data_weekly, outdir)
            if ts_objs_weekly_reg != None:
                for ts_data_weekly_reg in ts_objs_weekly_reg:
                    # writes json file
                    _write_stationdata_json(ts_data_weekly_reg, outdir)

        logger.info(
            f"Finished computing json files for {model_name} ({model_var}) vs. "
            f"{obs_name} ({obs_var})"
        )

        dt = time() - t00
        logger.info(f"Time expired (TOTAL): {dt:.2f} s")
